using System;
using System.ServiceModel;
using System.ServiceModel.Channels;
using System.ServiceModel.Description;
using System.ServiceModel.Dispatcher;
using System.Xml;
using System.Xml.Schema;

namespace innoQ.WCF.Extensions.Validation {
    /// <summary>
    /// Message Inspector zur Validierung gegen das im Konstruktor uebergebene (Message) Schema
    /// </summary>
    public class XsdValidationInspector : IDispatchMessageInspector {

        #region Properties
        private XmlSchemaSet schemas;
        /// <summary>
        /// Schemas, gegen die die Nachricht und das Dokument validiert werden
        /// </summary>
        protected XmlSchemaSet Schemas {
            get { return schemas; }
            set { schemas = value; }
        }
        
        private XmlReaderSettings readerSettings;
        /// <summary>
        /// Reader/Validation Settings fuer die XML Schema Validierung
        /// </summary>
        public XmlReaderSettings ReaderSettings {
            get {
                if (readerSettings == null) {
                    readerSettings = new XmlReaderSettings();
                    readerSettings.Schemas.Add(schemas);
                    readerSettings.ValidationType = ValidationType.Schema;
                    readerSettings.ConformanceLevel = ConformanceLevel.Auto;
                    readerSettings.ValidationFlags = XmlSchemaValidationFlags.ReportValidationWarnings;
                }
                return readerSettings;
            }
        }
        #endregion

        public XsdValidationInspector(XmlSchemaSet schemas) {
            this.schemas = schemas;
        }

        #region IDispatchMessageInspector Members
        public object AfterReceiveRequest(ref Message request, IClientChannel channel, InstanceContext instanceContext) {
            try {
                // Message sichern und Kopie erzeugen
                MessageBuffer buffer = request.CreateBufferedCopy(int.MaxValue);
                Message temp = buffer.CreateMessage();

                // Message content validieren
                XmlDocument bodyDoc = new XmlDocument();
                bodyDoc.Load(temp.GetReaderAtBodyContents());
                XmlReader r = XmlReader.Create(new XmlNodeReader(bodyDoc), ReaderSettings);
                while (r.Read()) ; // nur validieren

                // Neue Nachricht erzeugen und alte ersetzen
                request = buffer.CreateMessage();
            } catch (XmlSchemaException e) {
                throw new FaultException(e.Message);
            }
            return null;
        }

        public void BeforeSendReply(ref Message reply, object correlationState) {}
        #endregion

    }

    /// <summary>
    /// EndpointBehavior fuer Schema-Validierung.
    /// Fuer das Behavior muessen die Uri und der Namespace des Service-Endpoint-Message-Schemas konfiguriert werden.
    /// </summary>
    public class XsdValidationBehavior : IEndpointBehavior {
        #region Properties
        private string xsdUri;
        /// <summary>
        /// Uri des Message Schemas
        /// </summary>
        public string XsdUri {
            get { return xsdUri; }
            set { xsdUri = value; }
        }

        private string xsdTargetNamespace;
        /// <summary>
        /// Namespace des Message Schemas
        /// </summary>
        public string XsdTargetNamespace {
            get { return xsdTargetNamespace; }
            set { xsdTargetNamespace = value; }
        }
        #endregion

        public XsdValidationBehavior(string targetNamespace, string anUri) {
            XsdUri = anUri;
            XsdTargetNamespace = targetNamespace;
        }

        #region IEndpointBehavior Members
        public void AddBindingParameters(ServiceEndpoint serviceEndpoint, BindingParameterCollection bindingParameters) {
            foreach (object p in bindingParameters) {
                Console.WriteLine(p);
            }
        }

        public void ApplyClientBehavior(ServiceEndpoint serviceEndpoint, ClientRuntime behavior) {}

        public void ApplyDispatchBehavior(ServiceEndpoint serviceEndpoint, EndpointDispatcher endpointDispatcher) {
            XmlSchemaSet schemaSet = new XmlSchemaSet();
            schemaSet.Add(XsdTargetNamespace, XsdUri);
            schemaSet.Compile();
            endpointDispatcher.DispatchRuntime.MessageInspectors.Add(
                new XsdValidationInspector(schemaSet));
        }

        public void Validate(ServiceEndpoint serviceEndpoint) {}
        #endregion
    }
}